/*
***************************************************************************
  A 1024 (16x8x8) LED matrix WiFi message board with rotary encoder,
  BME280 sensor and a web interface with new message notification
  by ericBcreator

  It connects to Wifi and a timeserver and then displays date, time,
  sunrise and sunset for a set location, sensors reading of the 
  temperature, humidity and air pressure, current weather and forecasts 
  from openweather.map for multiple cities, top headlines from 
  NewsApi.org, stock data from Alpha Vantage and random Wikipedia items.
  
  A rotary encoder selects between the default display 'carousel',
  one news category, stock or weather data, no messages or it resets 
  the ESP.

  There is web access (enter the IP address of the ESP32 which is 
  displayed at startup in a browser) for displaying custom texts and 
  access to some controls. A buzzer plays a notification sound whenever 
  a new message is received.
***************************************************************************
  components:
  - ESP32
  - 16 x 8x8 MAX7219 LED matrices
  - BME280 sensor
  - rotary encoder
  - buzzer
  - light sensor
***************************************************************************
 last updated 20190411 by ericBcreator
  20181231: solved Raeren Openweather error
  20190126: added Youtube api support
  20190202: added IP to menu and menuItem numbers
  20190328: fixed webinterface for menuItem numbers
  20190406: added na_maxCallsPerDay for newsapi
            added clickable urls for newsapi and wikipedia for the webserver
  20190410: added light sensor support for setting the brightness of the leds
  
  TODO: adding buffering for newsapi
***************************************************************************

  This code is free for personal use, not for commercial purposes.
  Please leave this header intact.

  contact: ericBcreator@gmail.com
***************************************************************************
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*Brett Oliver's notes 01/05/2020
*Modification/additions to ericBcreator's code
*to get working had to roll back Json library to v5
* and used Dusk to daswm library from https://www.hackster.io/ericBcreator/1024-led-matrix-wifi-message-board-with-menu-web-interface-1b2666
* v04 added float sea_level- this is a correction to the air pressure reading for sea level correction
* v05 added version number to web interface and added print of pressre from sensor not corrected to sea level
* v06 add light sensors off pin 34
* v07 added choices for diff MAX7219 hardware settings and change rotary encoder pins
* v08 removed WIKI removed menu line 314 and remove and change case line 515
* v09 updated Kenley location
* v10 You Tube daily stats added 1423 
* v11 Ave hourly You Tube stats added 1470 
* v12 day length increase decrease added
* v15 sunrise sunset times added for yesterday
* v16 daylength added
* v18 daylength completed
* v19 added PIR controlled display blanking see PIRdetector 
* v21 correction for YouTube stats not working.
* v22 web API checking deactivated if PIR has not detected movement (does not include YouTube)
* 
* version number located at line 979 and 407
* 
* contact brett.oliver@btinternet.com
* brettoliver.org.uk
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#define FN_ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))  // Generic macro for obtaining number of elements of an array

//#define DEBUG                                 // enable debugging messages
//#define DEBUG_FAST                            // skip wifi/timeserver connect messages
//#define DEBUG_NEWSAPI_NC                      // do not connect to Newsapi.org (to save the number of daily requests)

#ifdef DEBUG
  #define DEBUGPRINT(x)   Serial.print(x)
  #define DEBUGPRINTLN(x) Serial.println(x)
#else
  #define DEBUGPRINT(x)
  #define DEBUGPRINTLN(x)
#endif

//#define LOCAL_LANG                              // @EB-setup use local language (user-definable, otherwise use English)
#define CONVERT_UTF8                            // @EB-setup converts UTF8 'special' characters to the Parola font set
#define WEBSERVER                               // @EB-setup enable webserver and web access
#define BUZZER                                  // @EB-setup use buzzer for new messages from the webserver

//
// Include libraries
//

#include <Adafruit_BME280.h>                    // BME280 library
#include <ArduinoJson.h>                        // Arduino Json library
#include <Dusk2Dawn.h>                          // Dusk2Dawn sunrise sunset library
#include <MD_MAX72xx.h>                         // Needed for Parola
#include <MD_Parola.h>                          // Parola LED library
#include <RotaryEncoder.h>                      // Rotary encoder library
#include <SPI.h>                                // Needed for the radiohead library
#include <TimeLib.h>                            // Time library
#include <WiFiClientSecure.h>                   // WiFi library
#include <WiFiUdp.h>                            // Udp library
#include <YoutubeApi.h>                         // YouTube library

#include "Font_Data_Numeric.h"                  // local file: font data for time

#ifdef CONVERT_UTF8
  #include "Font_Data_UTF8.h"                   // local file: font data for UTF8 support
#endif


//
// LED matrix setup and pins definition
//

//#define HARDWARE_TYPE MD_MAX72XX::FC16_HW       // @EB-setup define the type of hardware connected
//#define HARDWARE_TYPE MD_MAX72XX::PAROLA_HW      // @EB-setup define the type of hardware connected
//#define HARDWARE_TYPE MD_MAX72XX::GENERIC_HW      // @EB-setup define the type of hardware connected
#define HARDWARE_TYPE MD_MAX72XX::ICSTATION_HW      // @EB-setup define the type of hardware connected
#define CLK_PIN     18                          // @EB-setup CLK / SCK pin number  (EB setup: yellow)
#define DATA_PIN    23                          // @EB-setup DIN / MOSI pin number (EB setup: brown)
#define CS_PIN      5                           // @EB-setup LOAD / SS pin number  (EB setup: green)
#define MAX_DEVICES 16                          // @EB-setup num of LED matrices
#define MAX_ZONES   3                           // num of zones
MD_Parola PD = MD_Parola(HARDWARE_TYPE, CS_PIN, MAX_DEVICES);

//
// BME280 setup
//

Adafruit_BME280 bme;                            // connect via I2C pin 21 SDA and 22 SCL
uint8_t bme_address = 0x76;                     // @EB-setup set to 0x76 or 0x77
float sea_level = 16.68;                          //@EB-setup Brett sea level correction value for your location

//
// Rotary encoder setup
//

#define RE_PINA 14                              // Brett was 16 @EB-setup pinA / clock pin number
#define RE_PINB 27                              // Brett was 17 @EB-setup pinB / data pin number
#define RE_SWITCH 4                             // @EB-setup switch / switch pin number
RotaryEncoder RE_encoder(RE_PINA, RE_PINB);     // initialize the rotary encoder
int reLastPos = 0;
int reLastSwitchState = 0;

//
// Buzzer setup
//

#define BUZZER_PIN 15                           // @EB-setup buzzer pin
int buzz_volume = 100;                           // @EB-share buzzer volume (duty cycle 0-100)
int buzz_channel = 0;
int buzz_resolution = 8;

//
// Light sensor setup
//

#define LS_ACTIVE                               // @EB-setup comment this line when not using a light sensor (set the brightness with pd_brightness instead)
// LS_ACTIVE now enables PIR motionsensing
#define LS_PIN 34                               // @EB-setup light sensor pin

int ls_minValue = 0;                            // @EB-setup min brightness (0-15)
int ls_maxValue = 8;                           // @EB-setup max brightness (0-15)
int ls_delay = 500;                             // @EB-setup delay time for checking the light sensor (in milliseconds)
int ls_triggerStep = 1;                         // @EB-setup if value 1 causes flickering try 2 or 3

long ls_lastTime = 0;

//PIR Sensor  //Brett
int PIRPin = 25; // display is turned off when no motion from the optional PIR or Microwave sensor is detected



//
// Initialize variables
//

//Brett Yesterdays date sunrise sunset

byte Day=0;      // e.g. 28
byte Month=0;     // e.g. 8
int Year=0;    // e.g. 2017
int dayLength = 0; //length of day in minutes
int daymins = 0; // minutes part of day length
int dayhours = 0; // hours part of day length
int dayLengthYesterday = 0; //length of yesterday in minutes
int dayminsYesterday = 0; // minutes part of day length yesterday
int dayhoursYesterday = 0; // hours part of day length yesterday
 

tmElements_t tm;
int dayLengthRise2day = 0; //sunrise today mins after midnight
int dayLengthSet2day  = 0; //sunset today mins after midnight
int dayLengthRiseyesterday = 0; //sunrise yesterday mins after midnight
int dayLengthSetyesterday  = 0; //sunset yesterday mins after midnight
int dayIncreaseDecrease = 0; // daylength increase over a day




//end Brett Yesterdays date

/* WiFi connection */
const char * ssid     = "";                   // @EB-setup set your WiFi SSID was BTHub6-QCXR TP-LINK_AP_25AE EXT2-BTHub6-QCXR
const char * password = "";                   // @EB-setup set your WiFi password
int wifiTimeout = 30000;                                            // WiFi timeout (in milliseconds)

/* location for sunrise and sunset */
char d2d_City[] = "Kenley";                                      // @EB-share City name
float d2d_longitude = 51.32630;                                     // @EB-share
float d2d_lattitude = 0.10250;                                     // @EB-share
int d2d_timezone = 0;                                               // @EB-share UTC timezone offset

/* openweathermap connection */
const char * ow_servername = "api.openweathermap.org";              // remote server openweathermap.org
const int ow_port = 80;                                             // port
const String ow_APIKEY = "";        // @EB-setup personal api key for retrieving the weather data. leave empty to skip

#ifdef LOCAL_LANG
  const String ow_language = "NL";                                  // @EB-setup openweather language
#else
  const String ow_language = "EN";
#endif

int ow_cityIDLoop = 0;

// get the ID at https://openweathermap.org/ type the city, click search and click on the town
// then check the link, like this: https://openweathermap.org/city/5128581 5128581 is the ID for New York
// @EB-share 

String ow_cityIDs[] = {
  "2633993",      // Kenley(Whyteleafe)
  "2650613",       // Dursley
  "2647655",      // Hailsham
};

/* newsapi connection */
const char * na_servername = "newsapi.org";                         // remote server newsapi.org
const int na_port = 80;                                             // port
const int na_maxCallsPerDay = 1000;                                 // @EB-setup limit the calls per day. set to zero to disable
                                                                    // 1000 max calls per day for free developer plan;

const String na_APIKEY = "";        // @EB-setup personal api key. leave empty to skip

#ifdef LOCAL_LANG
  const String na_country = "NL";                                   // @EB-setup country code for news
#else
  const String na_country = "GB";// Brett was US
#endif

int na_article = 0;

/* alpha vantage stock connection */
const char * av_servername = "www.alphavantage.co";                 // remote server alphavantage
const int av_port = 443;                                            // port for secure connection

const String av_APIKEY = "";        // @EB-setup personal api key. leave empty to skip

// @EB-share
String av_symbolIDs[] = { "BTGOF" } ;
String av_symbolName[] = { "BT PLC USD" } ;

int av_symbolIDLoop = 0;

/* wikipedia connection */
const int wp_port = 443;                                            // port for secure connection

#ifdef LOCAL_LANG
  const char * wp_servername = "nl.wikipedia.org";                  // @EB-setup country code for wikipedia
#else
  const char * wp_servername = "en.wikipedia.org";
#endif

/* youtube connection */
const char * yt_servername = "www.googleapis.com";                  // remote server youtube
const int yt_port = 443;                                            // port for secure connection
const String yt_APIKEY = "";                                        // @EB-setup personal api key. leave empty to skip
const String yt_CHANNELID = "";     // @EB-setup personal channel ID

/* message definitions */
#ifdef LOCAL_LANG // @EB-share
  char msgWifiConnecting[] =                    "...wifi verbinden...";
  char msgWifiConnected[] =                     "verbonden met";
  char msgWifiFailed[] =                        "* wifi verbinding mislukt *";
  char msgConnectToTimeserver[] =               "...verbinden met de tijdserver...";
  char msgTimeserverFailed[] =                  "* verbinding met tijdserver mislukt *";
  char msgDSTon[] =                             "verbonden,   zomertijd  ";
  char msgDSToff[] =                            "verbonden,  wintertijd  ";  
  char msgResetting[] =                         "...auto reset...";
  char msgOpenweatherFailed[] =                 "* verbinding met Openweathermap mislukt *";
  char msgNewsapiFailed[] =                     "* verbinding met Newsapi.org mislukt *";
  char msgNewsapiError[] =                      "* fout bij ophalen van nieuws van Newsapi.org *";
  char msgAlphaVantageFailed[] =                "* verbinding met AlphaVantage.co mislukt *";
  char msgWikipediaFailed[] =                   "* Wikipedia *"; 
  
  char msgNews[] =                              "Laatste nieuws van";
  char msgSunrise[] =                           "Zonsopkomst";
  char msgSunset[] =                            "Zonsondergang";
  char msgCurrentWeather[] =                    "Het weer in";
  char msgForecast[] =                          "Weersverwachting voor";
  char msgTemp[] =                              "temperatuur";
  char msgHumidity[] =                          "vochtigheid";
  char msgClouds[] =                            "bewolking";
  char msgRain[] =                              "neerslag";
  char msgWind[] =                              "wind";
  char msgAVPrevious[] =                        "vorige slotkoers";
  char msgAVCurrent[] =                         "laatste koers";
  
  char msgSelected[] =                          "OK";
#else
  char msgWifiConnecting[] =                    "...wifi connecting...";
  char msgWifiConnected[] =                     "connected to";
  char msgWifiFailed[] =                        "*failed connecting to wifi*";
  char msgConnectToTimeserver[] =               "...connecting to the timeserver...";
  char msgTimeserverFailed[] =                  "* failed connecting to the timeserver *";
  char msgDSTon[] =                             "connected, DST active";
  char msgDSToff[] =                            "connected, no DST";
  char msgResetting[] =                         "...auto reset...";
  char msgOpenweatherFailed[] =                 "* failed connecting to Openweathermap *";
  char msgNewsapiFailed[] =                     "* failed connecting to Newsapi.org *";
  char msgNewsapiError[] =                      "* failed retrieving news from Newsapi.org ";
  char msgAlphaVantageFailed[] =                "* failed connecting to AlphaVantage.co *";
  char msgWikipediaFailed[] =                   "* failed connecting to Wikipedia *";
  char msgYoutubeFailed[] =                     "* failed connecting to YouTube *"; 
  
  char msgNews[] =                              "Latest news from";
  char msgSunrise[] =                           "Sunrise today";
  char msgYesterdaySunrise[] =                  "Sunrise yesterday";  //Brett for yesterdays sunrise/sunset
  char msgSunset[] =                            "Sunset today";        
  char msgYesterdaySunset[] =                   "Sunset yesterday"; //Brett for yesterdays sunrise/sunset
  char msgDaylight[] =                          "Day lenght"; //Brett for daylight hours
  char msghour[] =                              "hours"; //Brett for daylight hours
  char msgmin[] =                               "mins"; //Brett for daylight hours
  char msgshorter[] =                           "down"; //Brett for daylight hours
  char msglonger[] =                            "up"; //Brett for daylight hours
  char msgnochange[] =                          "no change"; //Brett for daylight hours
  char msgCurrentWeather[] =                    "Current weather in";
  char msgForecast[] =                          "Weather forecast for";
  char msgTemp[] =                              "temperature";
  char msgHumidity[] =                          "humidity";
  char msgClouds[] =                            "clouds";
  char msgRain[] =                              "rain";
  char msgWind[] =                              "wind";
  char msgAVPrevious[] =                        "previous close";
  char msgAVCurrent[] =                         "latest";
  
  char msgSelected[] =                          "OK";
#endif

/* menu setup */
// @EB-share
#ifdef LOCAL_LANG
  const String menu[] { 
    "Standaard mozaek", 
    "Nieuws headlines", 
    "Economie", 
    "Technologie", 
    "Algemeen", 
    "Entertainment", 
    "Gezondheid", 
    "Wetenschap", 
    "Sport", 
    "Wikipedia willekeurig item",
    "Beursinfo van Alpha Vantage", 
    "Weerberichten", 
    "Youtube", 
    "Eigen bericht", 
    "IP adres", 
    "Geen berichten", 
    "- Reset" 
  } ;
#else
  const String menu[] { 
    "Default carousel",                         //  0
    "Top headlines",                            //  1
    "Business",                                 //  2
    "Technology",                               //  3
    "General",                                  //  4
    "Entertainment",                            //  5
    "Health",                                   //  6
    "Science",                                  //  7
    "Sports",                                   //  8
    "Stock information from Alpha Vantage",     // 9
    "Weather information",                      // 10
    "Youtube",                                  // 11
    "User defined message",                     // 12
    "IP Address",                               // 13
    "No messages",                              // 14
    "- Reset"                                   // 15
  } ;
#endif

#define menuItemDefault    0
#define menuItemNews       1                    // 2 to 8 are news categories
#define menuItemStock     9 
#define menuItemWeather   10
#define menuItemYouTube   11
#define menuItemUser      12
#define menuItemIP        13
#define menuItemNoMess    14
#define menuItemReset     15

const String menuNewsApi[] { "", "&category=business", "&category=technology", "&category=general", "&category=entertainment", "&category=health", "&category=science", "&category=sports" } ;

int menuItem0 = menuItemDefault;                // @EB-share default menu selection
int menuReturnDelay = 10000;                    // @EB-share delay before the menu exits

int prevMenuItem0 = menuItem0;
int menuLen = FN_ARRAY_SIZE(menu);
int menuNewsCat = 0;

/* time server */
int timeServerDelay = 1000;                     // delay for the time server to reply
int timeServerPasses = 4;                       // number of tries to connect to the time server before timing out
boolean timeServerConnected = false;            // is set to true when the time is read from the server

unsigned int localPort = 2390;                  // local port to listen for UDP packets
IPAddress timeServerIP;                         // IP address of random server
const char* ntpServerName = "pool.ntp.org";     // server pool
byte packetBuffer[48];                          // buffer to hold incoming and outgoing packets
int timeZoneoffsetGMT = d2d_timezone * 3600;    // offset from Greenwich Meridan Time
boolean DST = false;                            // daylight saving time
WiFiUDP clockUDP;                               // initialize a UDP instance


//Brett
/* YouTube */
long yt_views_start = 0; // YouTube daily views at startup
long yt_views_today = 0; // YouTube views daily
long yt_views_hrsstart = 0; // YouTube hourly views at startup
long yt_views_hour = 0; // YouTube views last hour

int yt_views_date = 0; // date count started
int yt_prev_date = 0; // previous date
int yt_views_starthour = 0; // hour count started
int yt_prev_hour = 0; // previous hour
int yt_ave_views_hour = 0.0; //ave views per hour
int yt_start_hour = 0;  // 1sr day start hour compensation value for ave views per hour
int firstrun = 0; // firstrun detector for ave views per hour



//End Brett


/* web server */
WiFiServer webServer(80);
int webpageRefreshTime = 15;                    // @EB-share delay time before the webserver page will auto refresh

/* Parola */
#define BUF_SIZE 250                            // @EB-share buffer size for zone 2

char pd_message0[9] = { "" };                   // message buffer zone 0 (time)
char pd_message1[20] = { "" };                  // message buffer zone 1
char pd_message2[BUF_SIZE] = { "" };            // message buffer zone 2
char tmpBuffer[BUF_SIZE] = { "" };

char pd_messageUser[BUF_SIZE] = { "--- 1024 LED Matrix WiFi Information Ticker Board -- Brett Oliver 2020 v22.0 ---" }; // @EB-share

char pd_url[BUF_SIZE] = { "" };                 // url (for newsapi and wikipedia)

uint8_t degC[] = { 6, 3, 3, 56, 68, 68, 68 };   // Deg C character
uint8_t degF[] = { 6, 3, 3, 124, 20, 20, 4 };   // Deg F character
uint8_t degCascii = 0x8F;
uint8_t degFascii = 0x90;

int pd_brightness = 2;                          // @EB-share brightness (0-15)
int scrollSpeedSetup = 20;                      // @EB-share frame delay setup
int scrollSpeedMain = 27;                       // @EB-share frame delay main loop //Brett was 40
int scrollPause = 1500;                         // @EB-share delay after displaying the message (in milliseconds)
textEffect_t scrollEffect = PA_SCROLL_LEFT;     // @EB-share default effect
textPosition_t scrollAlign = PA_CENTER;         // @EB-share default align
textEffect_t zone1ScrollEffect = PA_SCROLL_UP;  // @EB-share default effect
int zone1Pause = 8000;   //Brett was 10000 delay for temp,hum,date pressure    // @EB-share delay after displaying the zone 1 message (in milliseconds)

/* general */
boolean tempCelsius = true;                     // @EB-setup set false for Fahrenheit

boolean colonFlag = true;
static uint8_t loopCounter1 = 0;
static uint8_t loopCounter2 = 0;
static uint32_t lastZone0Time = 0;
static uint32_t lastZone1Time = 0;

//
// Setup
//

void setup() {
  //Brett
yt_views_today = 0; //Brett sets days views to zero on startup
yt_views_hour = 0; // Sets hours views to zero on startup
yt_prev_date = 0; // previous date set to zero on startup
yt_prev_hour = 25; // previous hour set to zero onstartup


// end Brett
 PD.displayShutdown(false);// Brett enable displays
  
#ifdef DEBUG
  Serial.begin(57600);                          // for debugging
  Serial.println("");
  Serial.println("Setup...");
#endif

#ifdef BUZZER  
  ledcSetup(buzz_channel, 0, buzz_resolution);  // setup the buzzer
#endif

  pinMode(RE_SWITCH, INPUT_PULLUP);             // setup the encoder button as input high
  pinMode (PIRPin, INPUT_PULLDOWN);                      // held low unless Motion detected  //Brett

#ifdef LS_ACTIVE
 // pinMode(LS_PIN, INPUT_PULLDOWN);              // setup the light sensor input as input low 
  checkLightSensor();
 
#endif

 

  setupPD();                                    // start the Parola display

  bme.begin(bme_address);                       // start the BME280 sensor

  connectToWifi();                              // connect to Wifi

  if (WiFi.status() == WL_CONNECTED) {
    getTimeFromServer();                        // connect to the timeserver and get the local time

#ifdef WEBSERVER
    webServer.begin();
    #ifdef DEBUG
      Serial.println("Webserver started");
    #endif
#endif
  }
  
  PD.setSpeed(2, scrollSpeedMain);

 
}

//
// Main loop
//

void loop() { 
  PIRdetector (); // Brett sense movement if PIR or Microwave detector is connected
 //Brett resets start hour at midnight
  if(hour()== 0 && minute() == 0 && second() == 0) // resets yt_start_hour at 0:0:0 midnight 
    {
    yt_start_hour = 0;
   // Serial.println("yt_start_hour zeroed");
    }
    //end Brett
    
  if ((millis() - lastZone0Time >= 1000)) {                                         // wait for 1 second before displaying the time
    displayTime();
  
    lastZone0Time = millis();
  }

  if ((millis() - lastZone1Time >= zone1Pause)) {                                   // display the zone 1 animation after a pause
    lastZone1Time = millis();

    switch (loopCounter1) {
      case 0:     displayDateShort();           loopCounter1++;   break;            // display short date
      case 1:     displayTemp();                loopCounter1++;   break;            // display temperature
      case 2:     displayHumidity();            loopCounter1++;   break;            // display humidity
      case 3:     displayPressure();            loopCounter1 = 0; break;            // display pressure
    }
  }

  if (readRotEnc() != 0) {                                                          // if the rotary encoder is turned go to the menu
    selectMenu0();
    if (menuItem0 == menuItemReset)                                                 // reset the ESP
      ESP.restart();
      
    if (menuItem0 != prevMenuItem0) {
      na_article = 0;
      menuNewsCat = 0;
      loopCounter2 = 0;
      prevMenuItem0 = menuItem0;
    }
  } else if (readRotEncSwitch() == true)                                            // if the rotary encoder is pushed go to the next item to display
    resetDisplayZone2();

#ifdef WEBSERVER
  checkWebAccess();                                                                 // display the website and check for input
#endif

  if (WiFi.status() != WL_CONNECTED) {                                              // if there is no wifi connection reset after 2 minutes to try to reconnect
    
       
    if (millis() > (2 * 60000)) {
      PD.setPause(2, 0);
      displayMessage(msgResetting);
      ESP.restart();
    }
  } else if (menuItem0 != menuItemNoMess) {	                                        // only display if menu item 'No messages' is not selected
    if (PD.getZoneStatus(2)) {                                                      // display the next item when the animation of the current item has finished
      sprintf(pd_url, "");                                                          // reset the url

#ifdef DEBUG
      if (loopCounter2 == 0) {
        Serial.println("");
        Serial.println(pd_message0);
      }
#endif

      if (menuItem0 == menuItemDefault) {                                           // default setting, loop trough all items
        switch (loopCounter2) {
          case 0:     displayDate();            loopCounter2++; break;              // display date
          case 1:     displaySunrise();         loopCounter2++; break;              // display sunrise time     
          case 2:     displaySunset();          loopCounter2++; break;              // display sunset time
          case 3:     displayDaylength();       loopCounter2++; break;              // display daylength
          case 4:     displayWeatherData();     loopCounter2++; break;              // 1st loop current weather, 2nd loop weather forecast
          case 5:     displayNewsapiData();     loopCounter2++; break;              // display newsapi items
          case 6:     displayAVStockData();     loopCounter2++; break;              // display alpha vantage stock data
          case 7:     displayYT();              loopCounter2++; break;              // display YouTube  //Brett changed to 6        
          default:    loopCounter2 = 0;
        }

      } else {
          if (loopCounter2 == 10)
            displayDate();
          else {
            switch (menuItem0) {
              case menuItemUser:      displayUserMessage();  break;                 // user defined message (or posted on the webserver)
              case menuItemIP:        displayIP();           break;                 // only show IP
              case menuItemYouTube:   displayYT();           break;                 // only show YouTube
              case menuItemWeather:   displayWeatherData();  break;                 // only show weather data
              case menuItemStock:     displayAVStockData();  break;                 // only show stock data
          
              
              default:                                                              // only show news                  
                menuNewsCat = menuItem0 - 1;
                displayNewsapiData();
            }
          }
        loopCounter2 = (++loopCounter2) % 11;                                       // show 10 items and then display the date again
      }
      
#ifdef CONVERT_UTF8
      if (menuItem0 != menuItemUser)                                                // don't do a double conversion of the webserver text, it will erase the UTF characters
        utf8AsciiStr(pd_message2).toCharArray(pd_message2, BUF_SIZE);               // convert UTF8 characters
#endif
    }
  }

  PD.displayAnimate();

#ifdef LS_ACTIVE
  if (millis() - ls_lastTime > ls_delay) {                                         // check the light sensor every quarter second
    checkLightSensor();
   // PIRdetector (); // Brett sense movement if PIR or Microwave detector is connected
    ls_lastTime = millis();
  }
#endif
}

//
// Display functions
//

void setupPD() {
  PD.begin(MAX_ZONES);                          // start the Parola display

  PD.setZone(0, 0, 3);                          // zone 0 time
  PD.setZone(1, 4, 7);                          // zone 1 date, temperature, humidity
  PD.setZone(2, 8, MAX_DEVICES - 1);            // zone 2 messages

  PD.setFont(0, numeric7Seg);                   // set numeric font for the time

#ifdef CONVERT_UTF8
  PD.setFont(2, ExtASCII);                      // Parola font set with UTF8 support
#endif

  PD.setIntensity(pd_brightness);               // brightness (0-15)
  PD.setInvert(0, false);                       // invert false (text on, background off)
  PD.setInvert(1, false);                       // invert false (text on, background off)
  PD.setInvert(2, false);                       // invert false (text on, background off)

  PD.displayZoneText(0, pd_message0, PA_CENTER, 0, 0, PA_PRINT, PA_NO_EFFECT);
  PD.displayZoneText(1, pd_message1, PA_CENTER, scrollSpeedMain, scrollPause, zone1ScrollEffect, PA_NO_EFFECT);
  PD.displayZoneText(2, pd_message2, scrollAlign, scrollSpeedSetup, scrollPause, scrollEffect, scrollEffect);

  PD.addChar(degCascii, degC);
  PD.addChar(degFascii, degF);
}

void PIRdetector() {

 if (digitalRead (PIRPin) == HIGH)
 {
  PD.displayShutdown(false);
   
 
 }
 else if (digitalRead (PIRPin) == LOW )
 {
 PD.displayShutdown(true);
 
 }
}


void checkLightSensor() {
  int tmpValue;

  tmpValue = analogRead(LS_PIN);
 
  
  tmpValue = map(tmpValue, 0, 4095, ls_minValue, ls_maxValue);

  if (abs(tmpValue - pd_brightness) >= ls_triggerStep) {
    if (tmpValue < pd_brightness)
      pd_brightness--;      
    else
      pd_brightness++;
    PD.setIntensity(pd_brightness);
      
#ifdef DEBUG
     Serial.print("Brightness valoue ");//Brett
     Serial.println(analogRead(LS_PIN));//Brett
    Serial.print("Brightness set to ");
    Serial.println(pd_brightness);  
#endif
  }
}

void displayMessage(char * message) {
  snprintf(pd_message2, sizeof(pd_message2), "%s", message);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif

#if !defined (DEBUG_FAST)
  PD.displayReset(2);
  while (!PD.getZoneStatus(2))
    PD.displayAnimate();
#endif
}

void displayTime() {
  time_t t = now();

  if (hour(t) == 3 && minute(t) == 0)           // at 03:00 get the current time from the timeserver for readjusting and rereading DST
    getTimeFromServer();

  snprintf(pd_message0, sizeof(pd_message0), "%02d%c%02d", hour(t), (colonFlag ? ':' : ' '), minute(t));
  PD.displayReset(0);

  colonFlag = !colonFlag;
}

void displayDateShort() {
  time_t t = now();
 
  String monthStr = monthAsStringShort(month(t));
  snprintf(pd_message1, sizeof(pd_message1), "%d %s", day(t), monthStr.c_str());
  PD.displayReset(1);
}

void displayTemp() {
  float bmeTemp = bme.readTemperature();
  if (tempCelsius)
    sprintf(pd_message1, "%2.1f %c", bmeTemp, degCascii);
  else {
    bmeTemp = (bmeTemp * 9 / 5) + 32;
    sprintf(pd_message1, "%2.1f %c", bmeTemp, 129);
  }

#ifdef DEBUG
  Serial.println(pd_message1);
#endif
  PD.displayReset(1);
}

void displayHumidity() {
  float bmeHumidity = bme.readHumidity();
  sprintf(pd_message1, "%2.0f%% rh", bmeHumidity);
  
#ifdef DEBUG
  Serial.println(pd_message1);
#endif
  PD.displayReset(1);
}

void displayPressure() {
  #ifdef DEBUG
  // shows pressure from sensor not adjusted for sea level
  Serial.print("Pressure not sea level "); //Brett
  Serial.println(bme.readPressure()/100);//Brett
  #endif
  
  float bmePressure = (bme.readPressure() / 100.0f) + sea_level; // Brett added correction for local air pressure

  
  sprintf(pd_message1, "%4.0f h", bmePressure); // hPa doesn't fit...
  
#ifdef DEBUG
  Serial.println(pd_message1);
#endif
  PD.displayReset(1);
}

void displayDate() {
  time_t t = now();
  String dayStr = dayAsString(weekday(t));
  String monthStr = monthAsString(month(t));
  sprintf(pd_message2, "%s %d %s %04d", dayStr.c_str(), day(t), monthStr.c_str(), year(t));
  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}



void displaySunrise() {
  
  time_t t = now();
    time_t nextMakeTime;
    char d2dPreSunrise[] = "00:00";
    
    Day = day();
    Month = month();
    Year = year();
    tm.Day = Day;
    tm.Month = Month;
    tm.Year = Year - 1970; // offset from 1970;
    nextMakeTime =  makeTime(tm) - 86400 ; //date - 86400 secs in a day
  char d2dSunrise[] = "00:00";

  Dusk2Dawn d2d(d2d_longitude, d2d_lattitude, d2d_timezone);
  Dusk2Dawn::min2str(d2dSunrise, d2d.sunrise(year(t), month(t), day(t), DST));
  Dusk2Dawn::min2str(d2dPreSunrise, d2d.sunrise(year(nextMakeTime), month(nextMakeTime), day(nextMakeTime), DST));
  sprintf(pd_message2, "%s : %s %s %s %s", d2d_City, msgYesterdaySunrise, d2dPreSunrise, msgSunrise, d2dSunrise  );
  PD.displayReset(2);

 

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}

void displaySunset() {
  time_t t = now();
  time_t nextMakeTime;  
    char d2dPreSunset[] = "00:00";
    Day = day();
    Month = month();
    Year = year();
    tm.Day = Day;
    tm.Month = Month;
    tm.Year = Year - 1970; // offset from 1970;
    nextMakeTime =  makeTime(tm) - 86400 ; //date - number of seconds. There are 86400 secs in a day so the day and time is 1 day back when nextmaketime is called
  char d2dSunset[] = "00:00";

  Dusk2Dawn d2d(d2d_longitude, d2d_lattitude, d2d_timezone);
  Dusk2Dawn::min2str(d2dSunset, d2d.sunset(year(t), month(t), day(t), DST));
  Dusk2Dawn::min2str(d2dPreSunset, d2d.sunset(year(nextMakeTime), month(nextMakeTime), day(nextMakeTime), DST));
  sprintf(pd_message2, "%s : %s %s %s %s", d2d_City, msgYesterdaySunset, d2dPreSunset, msgSunset, d2dSunset);
  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
//displayDaylength();// Brett test untill menu added
}

void displayDaylength() {
  time_t t = now();
  time_t nextMakeTime;
   Day = day();
    Month = month();
    Year = year();
    tm.Day = Day;
    tm.Month = Month;
    tm.Year = Year - 1970; // offset from 1970;
    nextMakeTime =  makeTime(tm) - 86400 ; //date - 86400 secs in a day
  Dusk2Dawn d2d(d2d_longitude, d2d_lattitude, d2d_timezone);
dayLengthRise2day = d2d.sunrise(year(t), month(t), day(t), DST);
dayLengthSet2day = d2d.sunset(year(t), month(t), day(t), DST);

dayLengthRiseyesterday = d2d.sunrise(year(nextMakeTime), month(nextMakeTime), day(nextMakeTime), DST);
dayLengthSetyesterday = d2d.sunset(year(nextMakeTime), month(nextMakeTime), day(nextMakeTime), DST);

 
dayLength = dayLengthSet2day-dayLengthRise2day;
daymins = dayLength%60;
dayhours = (dayLength-daymins)/60;

dayLengthYesterday = dayLengthSetyesterday-dayLengthRiseyesterday;
dayminsYesterday = dayLengthYesterday%60;
dayhoursYesterday = (dayLengthYesterday-dayminsYesterday)/60;

dayIncreaseDecrease = dayLength - dayLengthYesterday;



 if ( dayIncreaseDecrease > 0)
 {
   sprintf(pd_message2, "%s : %s %i %s %i %s %i %s %s", d2d_City, msgDaylight, dayhours, msghour, daymins, msgmin, dayIncreaseDecrease, msgmin, msglonger );
 }
 else if ( dayIncreaseDecrease < 0)
 {
  sprintf(pd_message2, "%s : %s %i %s %i %s %i %s %s", d2d_City, msgDaylight, dayhours, msghour, daymins, msgmin, dayIncreaseDecrease, msgmin, msgshorter );
 }
 else
 sprintf(pd_message2, "%s : %s %i %s %i %s %s", d2d_City, msgDaylight, dayhours, msghour, daymins, msgmin, msgnochange );
  PD.displayReset(2);
  #ifdef DEBUG
  Serial.println(pd_message2);
#endif
 
 #ifdef DEBUG
 Serial.print("dayLengthRise2day "); 
  Serial.println(dayLengthRise2day);  
  Serial.print("dayLengthSet2day "); 
  Serial.println(dayLengthSet2day);  

   Serial.print("dayhours "); 
  Serial.println(dayhours);  
  Serial.print("daymins "); 
  Serial.println(daymins);  

  Serial.print("dayLengthRiseyesterday "); 
  Serial.println(dayLengthRiseyesterday);  
  Serial.print("dayLengthSetyesterday "); 
  Serial.println(dayLengthSetyesterday);

Serial.print("dayhoursYesterday "); 
  Serial.println(dayhoursYesterday);  
  Serial.print("dayminsYesterday "); 
  Serial.println(dayminsYesterday);

  Serial.print("dayIncreaseDecrease "); 
  Serial.println(dayIncreaseDecrease);
  #endif
}




void displayUserMessage() {
  strcpy(pd_message2, pd_messageUser);
  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}

void displayIP() {
  sprintf(pd_message2, "%s (%d dBm)  %s ", msgWifiConnected, WiFi.RSSI(), WiFi.localIP().toString().c_str());
  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}

void resetDisplayZone2() {
  strcpy(pd_message2, "");
  PD.displayClear(2);
  PD.displayReset(2);
}

//
// menu functions
//

void selectMenu0() {
#ifdef DEBUG
  Serial.println("Entering menu");
#endif

  int reResult;
  int lastReResult;
  int menuStartTime = millis();
  bool reSwitchPressed = false;

//  resetDisplayZone2;  // @EB: () missing, no use?? check

  PD.displayZoneText(2, pd_message2, PA_LEFT, scrollSpeedSetup, scrollPause, PA_PRINT, PA_CLOSING);
  displayMenuItem0();

  do {
    reResult = readRotEnc();

    if (reResult < 0) {             // down
      --menuItem0;
      if (menuItem0 < 0)
        menuItem0 = FN_ARRAY_SIZE(menu) - 1;
      PD.displayZoneText(2, pd_message2, PA_LEFT, scrollSpeedSetup, scrollPause, PA_SCROLL_UP, PA_SCROLL_UP);

    } else if (reResult > 0) {      // up
      menuItem0 = (++menuItem0) % FN_ARRAY_SIZE(menu);
      PD.displayZoneText(2, pd_message2, PA_LEFT, scrollSpeedSetup, scrollPause, PA_SCROLL_DOWN, PA_SCROLL_DOWN);
    }

    if (reResult != 0) {
      menuStartTime = millis();
      displayMenuItem0();
      lastReResult = reResult;
    } else {
      if (PD.getZoneStatus(2)) {
        if (lastReResult < 0)
          PD.setTextEffect(2, PA_SCROLL_LEFT, PA_SCROLL_UP);
        else
          PD.setTextEffect(2, PA_SCROLL_LEFT, PA_SCROLL_DOWN);
        PD.displayReset(2);
      }
    }

    reSwitchPressed = readRotEncSwitch();

    if (millis() - lastZone0Time >= 1000) {
      displayTime();
      lastZone0Time = millis();
    }

    PD.displayAnimate();

  } while ((millis() - menuStartTime < menuReturnDelay) && reSwitchPressed == false);

  if (reSwitchPressed == true) {
    strcpy(pd_message2, msgSelected);
    PD.displayZoneText(2, pd_message2, PA_LEFT, scrollSpeedSetup, 1000, PA_PRINT, PA_NO_EFFECT);
    while (!PD.getZoneStatus(2))
      PD.displayAnimate();

#ifdef DEBUG
    Serial.println(pd_message2);
#endif
  } else
    menuItem0 = prevMenuItem0;

  resetDisplayZone2();
  PD.displayZoneText(2, pd_message2, scrollAlign, scrollSpeedMain, scrollPause, scrollEffect, scrollEffect);

#ifdef DEBUG
  Serial.println("Exiting menu");
#endif
}

void displayMenuItem0() {
  sprintf(pd_message2, "%2d %s", menuItem0, menu[menuItem0].c_str());

#ifdef CONVERT_UTF8
  utf8AsciiStr(pd_message2).toCharArray(pd_message2, BUF_SIZE);
#endif
 
  PD.displayReset(2);

#ifdef DEBUG
  Serial.print("Menu: ");
  Serial.println(pd_message2);
#endif
}

int readRotEnc() {
  int reCurrentPos;
  int reResult = 0;

  RE_encoder.tick();

  reCurrentPos = RE_encoder.getPosition();
  if (reCurrentPos < reLastPos)
    reResult = -1;
  else if (reCurrentPos > reLastPos)
    reResult = 1;

  reLastPos = reCurrentPos;
  return reResult;
}

bool readRotEncSwitch() {
  int reSwitchState = digitalRead(RE_SWITCH);
  if (reSwitchState == LOW && reLastSwitchState == HIGH) {
    reLastSwitchState = reSwitchState;
    return true;
  }
  reLastSwitchState = reSwitchState;
  return false;
}

//
// web server
//

void checkWebAccess() {
  String currentLine = "";
  char oldMessage[BUF_SIZE] = { "" };
  char c;
  int customText, endOfText;  
  WiFiClient client = webServer.available();

  if (!client) 
    return;
    
#ifdef DEBUG
  Serial.println("Webserver client connected");
#endif

  strcpy(oldMessage, pd_message2);
  
  for (int i = 0; i < BUF_SIZE; i++) {
    if (oldMessage[i] == '\0') break;
    if (oldMessage[i] == degCascii) oldMessage[i] = 'C';                    // replace Celsius symbol with C for web page
  }
  
  for (int i = 0; i < BUF_SIZE; i++) {
    if (oldMessage[i] == '\0') break;
    if (oldMessage[i] == degFascii) oldMessage[i] = 'F';                    // replace Fahrenheid symbol with F for page
  }
  
  while (client.connected()) {
    if (client.available()) {
      c = client.read();

      if (c == '\n') {
        
#ifdef DEBUG
  if (currentLine.indexOf("GET /") >= 0) {
    Serial.print("Webserver client request: ");
    Serial.println(currentLine);
  }
#endif

        // check the GET statemnt

        customText = currentLine.indexOf("GET /?customtext=");
        if (customText >= 0) {
          endOfText = currentLine.indexOf("HTTP") - 1;
          parseGet((customText + 17), endOfText, currentLine).toCharArray(pd_messageUser, BUF_SIZE);

#ifdef DEBUG
          Serial.print("Webserver custom text: ");
          Serial.println(pd_messageUser);
#endif
          strcpy(oldMessage, pd_messageUser);
          menuItem0 = menuItemUser;
          loopCounter2 = 0;
          resetDisplayZone2();
          
#ifdef BUZZER
          playBuzzer();
#endif

        } else if (currentLine.indexOf("GET /nextItem") >= 0)
          resetDisplayZone2();
        else if (currentLine.indexOf("GET /carousel") >= 0)
          menuItem0 = menuItemDefault;
        else if (currentLine.indexOf("GET /nomessages") >= 0) {
          resetDisplayZone2();
          menuItem0 = menuItemNoMess;
        } else if (currentLine.indexOf("GET /youtube") >= 0)
          menuItem0 = menuItemYouTube;
        else if (currentLine.indexOf("GET /weather") >= 0)
          menuItem0 = menuItemWeather;
        else if (currentLine.indexOf("GET /stock") >= 0)
          menuItem0 = menuItemStock;
       // else if (currentLine.indexOf("GET /wikipedia") >= 0)
       //   menuItem0 = menuItemWiki;
        else if (currentLine.indexOf("GET /reset") >= 0)
          ESP.restart();
      
        // display the web page
        if (currentLine.length() == 0) {
#ifdef DEBUG
  Serial.println("Displaying website");
#endif

          client.println("HTTP/1.1 200 OK");
          client.println("Content-type:text/html");
          client.println("Connection: close");
          client.println();
        
          // display the HTML web page
          client.println("<!DOCTYPE html><html>");
          client.println("<head>");
          client.println("<meta http-equiv=\"Content-Type\" content=\"text/html;charset=ISO-8859-1\">");
          client.print("<meta http-equiv=\"refresh\" content=\"");
          client.print(webpageRefreshTime);
          client.print(";http://");
          client.print(WiFi.localIP());
          client.println("\" >");
          client.println("<style>html { font-family: Helvetica; margin: 0px 25%; text-align: center; } body { background-color: #EEEEEE; }");
          client.println(".button { background-color: #4CAF50; border: none; color: white; padding: 6px 16px;");
          client.println("text-decoration: none; font-size: 14px; font-weight: bold; margin: 4px; cursor: pointer;}");
          client.println("</style>");
          client.println("</head>");
        
          client.println("<body>");
          client.println("<h2>LED Message Board Web Interface</h2>");
          client.println("<h4>by ericBcreator</h4><br>");
          client.println("<h4>modified 23/04/2020 by Brett Oliver</h4><br>");
          client.println("<h4>v22.0</h4><br>");
        
          client.println("<p>Last displayed: </p>");
          client.println("<font color=\"red\"><b><p>");

          if (strlen(pd_url) == 0)
            client.println(oldMessage);
          else {
            client.print("<a href=\"");
            client.print(pd_url);
            client.print("\" target=\"_blank\">");
            client.print(oldMessage);
            client.println("</a>");
          }
            
          client.println("</b></p></font><br>");
          
          client.println("<form action=\"/\" method=\"get\">");
          client.println("<br>Input your message to display and click 'Send'<br><br>");
          client.println("<INPUT type=\"text\" size=\"60\" name=\"customtext\"<br>");
          client.println("<INPUT type=\"submit\" value=\"Send\">");
          client.println("</form><br>");

          client.print("<p>");
          client.println("<a href=\"/nextItem\"><button class=\"button\">Next item</button></a>");
          client.println("<a href=\"/carousel\"><button class=\"button\">Default carousel</button></a>");
          client.println("<a href=\"/wikipedia\"><button class=\"button\">Wikipedia</button></a>");
          client.println("<a href=\"/weather\"><button class=\"button\">Weather</button></a>");
          client.println("<a href=\"/stock\"><button class=\"button\">Stock info</button></a>");
          client.print("</p><p>");
          client.println("<a href=\"/youtube\"><button class=\"button\">Youtube</button></a>");
          client.println("<a href=\"/\"><button class=\"button\">Refresh</button></a>");
          client.println("<a href=\"/nomessages\"><button class=\"button\">No messages</button></a>");
          client.println("<a href=\"/reset\"><button class=\"button\">Reset the ESP32</button></a>");
          client.print("<p>");
          
          client.println("</body></html>");
          client.println();

          break;
        } else
          currentLine = "";

      } else if (c != '\r')
        currentLine += c;
    }
  }
 
  client.stop();

#ifdef DEBUG
  Serial.println("Webserver client disconnected");
#endif
}

//
// newsapi.org data
//

void displayNewsapiData() {
  if (na_APIKEY == "" or digitalRead (PIRPin) == LOW)// Brett if PIR is not detecting movement Newsapi not checked)
    return;

  static int na_numOfArticles;
  String url;
  String result;
  WiFiClient client;  
  
  static uint32_t lastNaTime;
  static char na_prevArt[BUF_SIZE] = { "" };
  static char na_prevUrl[BUF_SIZE] = { "" };
  
  if (na_maxCallsPerDay > 0) {
    if (lastNaTime > 0)
      if ((millis() - lastNaTime) < (86400000 / na_maxCallsPerDay)) {                 // limit the number of calls per day
        strcpy(pd_message2, na_prevArt);        // display the last read article
        strcpy(pd_url, na_prevUrl);
        PD.displayReset(2);
#ifdef DEBUG
  Serial.println("Skipping new newsapi.org item for max calls per day");
#endif
        return;
      }
    lastNaTime = millis();
  }
  
#ifdef DEBUG_NEWSAPI_NC                         // while debugging display a test message so the request counter doesn't increase
  sprintf(pd_message2, "debugging test getting news from Newsapi.org");
  PD.displayReset(2);
  Serial.println(pd_message2);
  return;
#endif

  if (!client.connect(na_servername, na_port))
    return;

  ++na_article;                                 // increase article counter

  url = "/v2/top-headlines?country=" + na_country + "&pageSize=1&page=" + String(na_article) + menuNewsApi[menuNewsCat].c_str() + "&apiKey=" + na_APIKEY;
//  url = "/v2/top-headlines?country=" + na_country + "&pageSize=" + String(8) + menuNewsApi[menuNewsCat].c_str() + "&apiKey=" + na_APIKEY;
//  note1: missing articles numbers in result with pageSize > 1 -> parse error.
//  note2: pageSize > 8 -> esp crashes

  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + na_servername + "\r\n" + "Connection: close\r\n\r\n");

#ifdef DEBUG
  Serial.println(String("GET ") + url + " HTTP/1.1\r\n");
#endif

  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 1000) {
      client.stop();
      return;
    }
  }

  result = "";
  while (client.available())
    result = client.readStringUntil('\r');

  client.stop();
  result.replace('[', ' ');
  result.replace(']', ' ');

  char jsonArray [result.length() + 1];
  result.toCharArray(jsonArray, sizeof(jsonArray));
  jsonArray[result.length() + 1] = '\0';

  StaticJsonBuffer<1024> json_buf;
  JsonObject &root = json_buf.parseObject(jsonArray);

  if (!root.success())
    strcpy(pd_message2, msgNewsapiError);
  else {
    if (root["status"] == "error") {
      sprintf(pd_message2, msgNewsapiError);
    } else {      
      na_numOfArticles = root["totalResults"];

#ifdef DEBUG
  Serial.print("Displaying article ");
  Serial.print(na_article);
  Serial.print(" of ");
  Serial.print(na_numOfArticles);
  Serial.println(" articles ");
#endif
      
      String asou = root["articles"]["source"]["name"];
      String atit = root["articles"]["title"];
      String aurl = root["articles"]["url"];
    
      snprintf(pd_url, sizeof(pd_url), "%s", aurl.c_str());

      if (menuItem0 == 0)                                           // the 'latest news' intro when displaying news items only
        snprintf(pd_message2, sizeof(pd_message2), "%s %s : %s", msgNews, asou.c_str(), atit.c_str());
      else
        snprintf(pd_message2, sizeof(pd_message2), "%s : %s", asou.c_str(), atit.c_str());        
      
      na_article = na_article % na_numOfArticles;                   // reset when all articles have displayed

      strcpy(na_prevArt, pd_message2);                              // copy the message and url to display when skipping for max calls per day
      strcpy(na_prevUrl, pd_url);
    }
  }

  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}

//
// weatherdata
//

void displayWeatherData() {
  if (ow_APIKEY == "" or digitalRead (PIRPin) == LOW)//Brett if PIR is not detecting movement Weather not checked)
    return;

  static bool ow_current = true;
  String url;
  String result;
  WiFiClient client;

  if (!client.connect(ow_servername, ow_port))
    return;

  String cityID = ow_cityIDs[ow_cityIDLoop];

  if (ow_current)                               // get current weather check weather properties at https://openweathermap.org/current
    url = "/data/2.5/weather?id=";
  else                                          // get forecast check properties forecasts at https://openweathermap.org/forecast5
    url = "/data/2.5/forecast?id=";

  url = url + cityID + "&units=metric&cnt=1&lang=" + ow_language + "&APPID=" + ow_APIKEY;
  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + ow_servername + "\r\n" + "Connection: close\r\n\r\n");

#ifdef DEBUG
  Serial.println(url);
#endif

  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 1000) {
      client.stop();
      return;
    }
  }

  result = "";
  while (client.available())
    result = client.readStringUntil('\r');

  client.stop();

  if (!ow_current) {
    result.replace('[', ' ');
    result.replace(']', ' ');
  }
  
#ifdef DEBUG
  Serial.println(result);
#endif

  char jsonArray [result.length() + 1];
  result.toCharArray(jsonArray, sizeof(jsonArray));
  jsonArray[result.length() + 1] = '\0';

  StaticJsonBuffer<1024> json_buf;
  JsonObject &root = json_buf.parseObject(jsonArray);

  if (!root.success())
    strcpy(pd_message2, msgOpenweatherFailed);

  else {
    if (ow_current) {
      String wcit = root["name"];
      String wdes = root["weather"][0]["description"];
      float  wtem = root["main"]["temp"];
      float  whum = root["main"]["humidity"];
      float  wclo = root["clouds"]["all"];
      float  wrai = root["rain"]["3h"];
      float  wwsp = root["wind"]["speed"];
      int    wwde = root["wind"]["deg"];

      snprintf(pd_message2, sizeof(pd_message2), "%s %s : %s, %s %2.1f %c  %s %3.0f%%  %s %3.0f%%  %s %3.1f mm  %s %2.0f m/s %s", 
        msgCurrentWeather, wcit.c_str(), wdes.c_str(), msgTemp, wtem, degCascii, msgHumidity, whum, msgClouds, wclo, msgRain, wrai, msgWind, wwsp, getWindDirection(wwde).c_str());
    } else {
      String wcit = root["city"]["name"];
      String wdes = root["list"]["weather"]["description"];
      float  wtmi = root["list"]["main"]["temp_min"];
      float  wtma = root["list"]["main"]["temp_max"];
      float  whum = root["list"]["main"]["humidity"];
      float  wclo = root["list"]["clouds"]["all"];
      float  wrai = root["list"]["rain"]["3h"];
      float  wwsp = root["list"]["wind"]["speed"];
      int    wwde = root["list"]["wind"]["deg"];

      snprintf(pd_message2, sizeof(pd_message2), "%s %s : %s, %s %2.1f %c - %2.1f %c  %s %3.0f%%  %s %3.0f%%  %s %3.1f mm  %s %2.0f m/s %s", 
        msgForecast, wcit.c_str(), wdes.c_str(), msgTemp, wtmi, degCascii, wtma, degCascii, msgHumidity, whum, msgClouds, wclo, msgRain, wrai, msgWind, wwsp, getWindDirection(wwde).c_str());
    }
  }

  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif

  ow_cityIDLoop = (++ow_cityIDLoop) % FN_ARRAY_SIZE(ow_cityIDs);

  if (ow_cityIDLoop == 0)                       // after a loop of all cities, toggle between current weather/forecasts
    ow_current = !ow_current;
}

//
// alpha vantage stock data
//

void displayAVStockData() {
  if (av_APIKEY == "" or digitalRead (PIRPin) == LOW)//Brett if PIR is not detecting movement Stocks not checked)
    return;

  String url;
  String result = "";
  String line = "";

  WiFiClientSecure client;

  if (!client.connect(av_servername, av_port))
    return;

  String stockID = av_symbolIDs[av_symbolIDLoop];
  String stockName = av_symbolName[av_symbolIDLoop];

  url = "/query?function=GLOBAL_QUOTE&symbol=" + stockID + "&apikey=" + av_APIKEY;
  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + av_servername + "\r\n" + "Connection: close\r\n\r\n");

#ifdef DEBUG
  Serial.println(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + av_servername);
#endif

  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 1000) {
      client.stop();
      return;
    }
  }

  while (client.available()) {
    line = client.readStringUntil('\r');
    result += line;
  }
  client.stop();

  result = result.substring(result.indexOf('{'), result.lastIndexOf('}') + 1);    // remove the header
  result.replace('\n', ' ');                                                      // remove the newline characters

  result.replace('[', ' ');
  result.replace(']', ' ');

  char jsonArray [result.length() + 1];
  result.toCharArray(jsonArray, sizeof(jsonArray));
  jsonArray[result.length() + 1] = '\0';

  StaticJsonBuffer<1024> json_buf;
  JsonObject &root = json_buf.parseObject(jsonArray);

  if (!root.success())
    strcpy(pd_message2, msgAlphaVantageFailed);
  else {
    float scur = root["Global Quote"]["05. price"];
    float sprc = root["Global Quote"]["08. previous close"];
    float scha = root["Global Quote"]["09. change"];
    float sper = root["Global Quote"]["10. change percent"];

    String scch = "";
    String scpe = "";
    if (scha > 0)
      scch = "+";
    if (sper > 0)
      scpe = "+";

    sprintf(pd_message2, "%s : %s  %6.2f  %s  %6.2f  %s%4.2f  %s%2.2f%%", stockName.c_str(), msgAVPrevious, sprc, msgAVCurrent, scur, scch.c_str(), scha, scpe.c_str(), sper);
  }

  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif

  av_symbolIDLoop = (++av_symbolIDLoop ) % FN_ARRAY_SIZE(av_symbolIDs);
}

//
// Wikipedia
//

void displayWikipedia() {
  String url;
  String result;
  WiFiClientSecure client;

  if (!client.connect(wp_servername, wp_port))
    return;

//  url = "/w/api.php?format=json&utf8=&action=query&generator=random&grnnamespace=0&prop=extracts&exintro&explaintext&redirects=1&generator=random";
  url = "/w/api.php?format=json&utf8=&action=query&generator=random&grnnamespace=0&prop=info|extracts&inprop=url&exintro&explaintext&redirects=1&generator=random";

  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + wp_servername + "\r\n" + "Connection: close\r\n\r\n");

#ifdef DEBUG
  Serial.println(String("GET ") + url + " HTTP/1.1");
#endif

  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 1000) {
      client.stop();
      return;
    }
  }

  result = "";
  while (client.available())
    result = client.readStringUntil('\r');

  client.stop();
  result.replace('[', ' ');
  result.replace(']', ' ');

  // get the 1st pageID from the returned result, format is "pages":{"0000000":
  int pagesIX = result.indexOf("pages") + 9;
  int pageEndIX = result.indexOf("\"", pagesIX);
  String pageID = result.substring(pagesIX, pageEndIX);

  char jsonArray [result.length() + 1];
  result.toCharArray(jsonArray, sizeof(jsonArray));
  jsonArray[result.length() + 1] = '\0';

  StaticJsonBuffer<1024> json_buf;
  JsonObject &root = json_buf.parseObject(jsonArray);

  if (!root.success())
    strcpy(pd_message2, msgWikipediaFailed);
  else {
    String wext = root["query"]["pages"][pageID]["extract"];
    String wurl = root["query"]["pages"][pageID]["fullurl"];
    snprintf(pd_message2, sizeof(pd_message2), "Wikipedia %s : %s", pageID.c_str(), wext.c_str());
    snprintf(pd_url, sizeof(pd_url), "%s", wurl.c_str());
  }

  PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
}

//
// Youtube //Brett reporting API
//

void displayYT() {
  if (yt_APIKEY == "")
    return;

  String line = "";
  String url = "";
  String result;
  WiFiClientSecure client;

  client.flush();
  if (!client.connect(yt_servername, yt_port)) {
    return;
  }

  yield();
  
  url = "/youtube/v3/channels?part=statistics&id=" + yt_CHANNELID + "&key=" + yt_APIKEY;
  //client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + yt_servername + "\r\n Cache-Control: no-cache\r\n Connection: close\r\n\r\n");
  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + yt_servername + "\r\nCache-Control: no-cache\r\n\r\n");
  DEBUGPRINTLN("GET " + url);
  
  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 1000) {
      client.stop();
      return;
    }
  }

  while (client.available()) {
    line = client.readStringUntil('\r');
    result += line;
  }
  client.stop();

  result = result.substring(result.indexOf('{'), result.lastIndexOf('}') + 1);    // remove the header
  result.replace('\n', ' ');                                                      // remove the newline characters

  result.replace('[', ' ');
  result.replace(']', ' ');

  char jsonArray [result.length() + 1];
  result.toCharArray(jsonArray, sizeof(jsonArray));
  jsonArray[result.length() + 1] = '\0';

  StaticJsonBuffer<1024> json_buf;
  JsonObject &root = json_buf.parseObject(jsonArray);

  if (!root.success()) {
    strcpy(pd_message2, msgYoutubeFailed);
 return;
  }
//      root["items"]["statistics"]["subscriberCount"];
//      root["items"]["statistics"]["viewCount"];
//      root["items"]["statistics"]["commentCount"];
//      root["items"]["statistics"]["hiddenSubscriberCount"];
//      root["items"]["statistics"]["videoCount"]; 
    
  unsigned long yt_subs = root["items"]["statistics"]["subscriberCount"];
  unsigned long yt_views = root["items"]["statistics"]["viewCount"];
  #ifdef DEBUG
 //Brett
 Serial.print("Time ");
 Serial.print(hour());
 Serial.print(":");
 Serial.print(minute());
 Serial.print(":");
 Serial.println(second());
 #endif
 yt_views_date = day();//date for YouTube stats
 yt_views_starthour = hour();//hour for YouTube stats
 
// Sense date change & reset daily count to running total
 if (yt_prev_date != yt_views_date)
 {
 yt_prev_date = yt_views_date ;
 yt_views_start = yt_views;// resets the start of day count yt_views_start to the running total
 
 }

//Sense hour change & reset hour count to running total
 if (yt_prev_hour != yt_views_starthour)
 {
 yt_prev_hour = yt_views_starthour ;
 yt_views_hrsstart = yt_views;// resets the start of day count yt_views_start to the running total
 
 }

 

    
    yt_views_today =  yt_views - yt_views_start; //YouTube daily views takes views at start of day off of running total views
    yt_views_hour =  yt_views - yt_views_hrsstart; //YouTube hourly views takes views at start of hour off of running total views

    if(firstrun == 0)// runs once only
    {
    yt_start_hour = hour(); // sets 1st day hour compensation if clock is started after midnight
    firstrun = 1;
    }
    
    yt_ave_views_hour = yt_views_today/(hour()+1 - yt_start_hour); // +1 at midnite views hour = todays views. 1 oclock = 0:00 to 1:59 so = 2
   
   
    #ifdef DEBUG    
    Serial.print("yt start hour conpensation " );
    Serial.println(yt_start_hour);

     Serial.print("yt_ave_views_hour " );
    Serial.println(yt_ave_views_hour);
    #endif
  
  //end Brett  
    sprintf(pd_message2, "Youtube : subscribers  %d  views  %d  today  %d  this hour %d  average per hour %d", yt_subs, yt_views, yt_views_today, yt_views_hour, yt_ave_views_hour );
    PD.displayReset(2);

#ifdef DEBUG
  Serial.println(pd_message2);
#endif
  
}


//
// network functions
//

void connectToWifi() {
#ifdef DEBUG
  Serial.println(msgWifiConnecting);
#endif

  int startTime = millis();

  //  WiFi.enableSTA(true);
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  WiFi.begin(ssid, password);

#if !defined (DEBUG_FAST)
  PD.setPause(2, 0);
  strcpy(pd_message2, msgWifiConnecting);
  PD.displayReset(2);
#endif

  while (WiFi.status() != WL_CONNECTED) {
#if !defined (DEBUG_FAST)
    PD.displayAnimate();
    if (PD.getZoneStatus(2))
      PD.displayReset(2);
#endif
    if ((millis() - startTime) > wifiTimeout)
      break;
  }

#if !defined (DEBUG_FAST)
  while (!PD.getZoneStatus(2))
    PD.displayAnimate();

  PD.setPause(2, scrollPause);
#endif

  if (WiFi.status() == WL_CONNECTED) {
#ifdef DEBUG
    Serial.print(msgWifiConnected);
    Serial.print(" ");
    Serial.println(WiFi.localIP().toString().c_str());
#endif
#if !defined (DEBUG_FAST)
    sprintf(pd_message2, "%s (%d dBm)  %s ", msgWifiConnected, WiFi.RSSI(), WiFi.localIP().toString().c_str());
#endif
  } else {
#ifdef DEBUG
    Serial.println(msgWifiFailed);
#endif
#if !defined (DEBUG_FAST)
    strcpy(pd_message2, msgWifiFailed);
#endif
  }

#if !defined (DEBUG_FAST)
  PD.displayReset(2);
  while (!PD.getZoneStatus(2))
    PD.displayAnimate();
#endif
}

void getTimeFromServer() {
  int connectStatus = 0, i = 0;
  unsigned long unixTime;
  clockUDP.begin(localPort);

  while (i < timeServerPasses && !connectStatus) {
    displayMessage(msgConnectToTimeserver);

    WiFi.hostByName(ntpServerName, timeServerIP);
    sendNTPpacket(timeServerIP);
    delay(timeServerDelay / 2);
    connectStatus = clockUDP.parsePacket();
    delay(timeServerDelay / 2);
    i++;
  }

  if (connectStatus) {
    timeServerConnected = true;
    clockUDP.read(packetBuffer, 48);

    // the timestamp starts at byte 40 of the received packet and is four bytes, or two words, long.
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
    // the timestamp is in seconds from 1900, add 70 years to get Unixtime
    unixTime = (highWord << 16 | lowWord) - 2208988800 + timeZoneoffsetGMT;
    setTime(unixTime);

    checkDST();
    if (DST) {
      unixTime = unixTime + 3600;
      setTime(unixTime);
      displayMessage(msgDSTon);
    } else
      displayMessage(msgDSToff);
  } else
    displayMessage(msgTimeserverFailed);

  strcpy(pd_message2, "");
}

void sendNTPpacket(IPAddress& address) {
  memset(packetBuffer, 0, 48);
  packetBuffer[0] = 0b11100011;                 // LI, Version, Mode
  packetBuffer[1] = 0;                          // Stratum, or type of clock
  packetBuffer[2] = 6;                          // Polling Interval
  packetBuffer[3] = 0xEC;                       // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;

  clockUDP.beginPacket(address, 123);           // NTP requests are to port 123
  clockUDP.write(packetBuffer, 48);
  clockUDP.endPacket();
}

//
// date functions
//

String dayAsString(int dayNum) {
#ifdef LOCAL_LANG // @EB-share
  String days[] { "Zondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrijdag", "Zaterdag" } ;
#else
  String days[] { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" } ;
#endif

  if (dayNum >= 0 && dayNum <= 7)
    return days[dayNum - 1];
  else
    return "";
}

String monthAsStringShort(int monthNum) {
#ifdef LOCAL_LANG // @EB-share
  String months[] { "Jan", "Feb", "Maa", "Apr", "Mei", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dec" } ;
#else
  String months[] { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" } ;
#endif

  if (monthNum >= 0 && monthNum <= 12)
    return months[monthNum - 1];
  else
    return "";
}

String monthAsString(int monthNum) {
#ifdef LOCAL_LANG // @EB-share
  String months[] { "Januari", "Februari", "Maart", "April", "Mei", "Juni", "Juli", "Augustus", "September", "Oktober", "November", "December" } ;
#else
  String months[] { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" } ;
#endif

  if (monthNum >= 0 && monthNum <= 12)
    return months[monthNum - 1];
  else
    return "";
}

String getWindDirection (int degrees) {
  int sector = ((degrees + 11) / 22.5 - 1);

#ifdef LOCAL_LANG // @EB-share
  String windDir[] {
    "noord", "noord-noordoost", "noordoost", "oost-noordoost", "oost", "oost-zuidoost", "zuidoost", "zuid-zuidoost",
    "zuid", "zuid-zuidwest", "zuidwest", "west-zuidwest", "west", "west-noordwest", "noordwest", "noord-noordwest"
  };
#else
  String windDir[] {
    "north", "north-northeast", "northeast", "east-northeast", "east", "east-southeast", "southeast", "south-southeast",
    "south", "south-southwest", "southwest", "west-southwest", "west", "west-northwest", "northwest", "north-northwest"
  };
#endif

  return windDir[sector];
}

//
// other functions
//

void checkDST() {
  time_t t = now();
  int y = year(t) - 2000;                       // subtract 2000 from the year
  int x1 = (y + y / 4 - 2) % 7;                 // identifies Sunday offset for March
  int x2 = (y + y / 4 + 2) % 7;                 // identifies Sunday offset for October

  // DST: begins last Sunday of March 02:00
  if ((month(t) == 3) && (day(t) == (31 - x1)) && (hour(t) >= 2))
    DST = true;
  if ((month(t) == 3) && (day(t) > (31 - x1)) || (month(t)  > 3))
    DST = true;

  // DST: ends last Sunday of Oct 02:00
  if ((month(t) == 10) && (day(t) == (31 - x2)) && (hour(t) >= 2))
    DST = false;
  if ((month(t) == 10) && (day(t) > (31 - x2)) || (month(t) > 10) || (month(t) < 3))
    DST = false;
}

String parseGet (int startChar, int endChar, String getText) {
  uint8_t c, i, count = 0;
  char tmpChar[] = { "0x00" };

  for (i = startChar; i < endChar; i++) {
    c = getText.charAt(i);
    switch (c) {
      case 0x2B: c = 0x20; break;               // convert '+' to space
      case 0x25:                                // %
        if (i <= endChar - 2) {                 // check text length
          if (getText.charAt(i + 1) == 0x25)    // skip if double %
            c = '\0';
          else                                  // get next 2 characters and convert the hex value to dec
            tmpChar[2] = getText.charAt(i + 1);          
            tmpChar[3] = getText.charAt(i + 2);
            tmpChar[4] = '\0';
            c = (int) strtol(tmpChar, 0, 16);
            i = i + 2;
        }
    }
    if (c != '\0')
      tmpBuffer[count++] = c;    
  }
  
  tmpBuffer[count] = '\0';
  return(tmpBuffer);
}

void playBuzzer() {
#ifdef DEBUG
  Serial.println("Playing buzzer");
#endif

  ledcAttachPin(BUZZER_PIN, buzz_channel);  
  ledcWrite(buzz_channel, buzz_volume);
  for (int i = 600; i <= 1000; i += 200) {
    ledcWriteTone(buzz_channel, i);
    delay(100);
  }
  ledcWriteTone(buzz_channel, 0);
  ledcWrite(buzz_channel, LOW);
  ledcDetachPin(BUZZER_PIN);

  pinMode(BUZZER_PIN, INPUT);                   // make sure the buzzer is silent ;-)
}

//
// UTF8 conversion
//

uint8_t utf8Ascii(uint8_t ascii) {
  // Modified from original code at http://playground.arduino.cc/Main/Utf8ascii
  // Extended ASCII encoding should match the characters at http://www.ascii-code.com/  
  static uint8_t cPrev;
  uint8_t c = '\0';
  
  if (ascii < 0x7f || ascii == degCascii || ascii == degFascii) {
    cPrev = '\0';
    c = ascii;
  } else {
    switch (cPrev) {
      case 0xC2: c = ascii;  break;
      case 0xC3: c = ascii | 0xC0;  break;
      case 0x82: if (ascii==0xAC) c = 0x80;   // Euro symbol special case
    }
    cPrev = ascii;                            // save last char
  }
  return(c);
}

String utf8AsciiStr(char * s) {
  // Converting a UTF-8 string to Extended ASCII   
  uint8_t c, k = 0;
  
  while (*s != '\0') {
    c = utf8Ascii(*s++);
    if (c != '\0')
      tmpBuffer[k++] = c;
  }
  tmpBuffer[k] = '\0';
  return(tmpBuffer);
}
